/* @(#)parse.c	1.5 97/12/23 */

/*
 * Copyright (c) Communications Research Centre, Government of Canada 1997. 
 * All rights reserved.
 * 
 * License is granted to copy and use this software, for research and
 * evaluation purposes, provided that the Communications Research Centre
 * is acknowledged in all documentation pertaining to any such copy or
 * use. The Communications Research Centre grants no other licenses
 * expressed or implied. Title, ownership rights, and intellectual
 * property rights in the software shall remain with the Communications
 * Research Centre.  The Centre's name and the Government of Canada
 * should not be used in any advertising without written permission.
 * 
 * THE COMMUNICATIONS RESEARCH CENTRE MAKES NO REPRESENTATIONS CONCERNING
 * EITHER THE MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided
 * "as is" without express or implied warranty of any kind.
 * 
 * These notices must be retained in any copies of any part of this
 * software.
 */


#include "mpoll.h"
struct topic tmptopic;
struct response tmpresponse;
struct member tmpmember;
char *buff = NULL;
char *did = NULL;
char *dtheme = NULL;
char *daddress = NULL;
char *duser = NULL;

/* --------------------------------------------------------------- */
/* setup temporary space for parsing */
setparsevars()
   {
   int i;
   
   /* do this only once so there are no memory leaks */

   /* setup temporary topic structure */
   
   tmptopic.qid = (char *)malloc(TMPSTRLEN);
   tmptopic.theme = (char *)malloc(TMPSTRLEN);
   tmptopic.email = (char *)malloc(TMPSTRLEN);
   tmptopic.summary = (char *)malloc(TMPSTRLEN);
   tmptopic.question = (char *)malloc(QUESTION_SIZE);
   tmptopic.option = (char *)malloc(TMPSTRLEN);
   for (i = 1; i <= MAX_RESPONSES; i++)
      {
      tmptopic.response[i] = (char *)malloc(TMPSTRLEN);
      }
  
   /* setup temporary response structure */
   
   tmpresponse.id = (char *)malloc(TMPSTRLEN);
   tmpresponse.theme = (char *)malloc(TMPSTRLEN);
   tmpresponse.address = (char *)malloc(TMPSTRLEN);
   tmpresponse.user = (char *)malloc(TMPSTRLEN);
   for (i = 1; i <= MAX_RESPONSES; i++)
      {
      tmpresponse.response[i] = (char *)malloc(TMPSTRLEN);
      }

   /* setup temporary member structure */
   
   tmpmember.theme = (char *)malloc(TMPSTRLEN);
   tmpmember.user = (char *)malloc(TMPSTRLEN);
   tmpmember.address = (char *)malloc(TMPSTRLEN);
   tmpmember.host = (char *)malloc(TMPSTRLEN);
   tmpmember.os = (char *)malloc(TMPSTRLEN);
   tmpmember.version = (char *)malloc(TMPSTRLEN);
   
   /* and a few extra variables */
   
   buff = (char *)malloc(BUFFER_SIZE);  
   did = (char *)malloc(TMPSTRLEN);  
   dtheme = (char *)malloc(TMPSTRLEN);  
   daddress = (char *)malloc(TMPSTRLEN);  
   duser = (char *)malloc(TMPSTRLEN);  
  
   }

/* --------------------------------------------------------------- */
/* parse an incoming data packet and determine it's type */
parse_packet(char *data)
   {
   char wbuff[TMPSTRLEN];
   
   if (strstr(data, "TYPE:question"))
      {
      parse_question(data);
      }
   else if (strstr(data, "TYPE:delete_question"))
      {
      parse_delete_question(data);
      }
   else if (strstr(data, "TYPE:response"))
      {
      parse_response(data);
      }
   else if (strstr(data, "TYPE:member"))
      {
      parse_member(data);
      }
   else 
      {
      sprintf(wbuff, 
         "Could not recognize the type of this data packet \n{%s}\n", 
         data);
      warning(wbuff);
      }
   }


/* ----------------------------------------------------------------- */
/* parse an incoming response packet */
parse_response(char *data)
   {
   char  *strafter();
   char *tok = NULL;
   char tmpl[1024];
   char *tmpp = tmpl;
   int i, j, t, targettopic;
   unsigned long remoteip;
   int numr, totalr;
   time_t now, ntpnow;
   long int tdiff;
   
   numr = 0; totalr = 0;
   
   /* set the tmp data structure to nothing ? */
   strcpy(tmpresponse.id, "");
   strcpy(tmpresponse.theme, "");
   tmpresponse.version = 0;
   strcpy(tmpresponse.user, "");
   strcpy(tmpresponse.address, "");
   for (j = 1; j <= MAX_RESPONSES; j++)
      {
      strcpy(tmpresponse.response[j], "");
      }
   tmpresponse.time = 0;
   
   
   strcpy(buff, data); /* set buff to the beginning */
   tok = strtok(buff, "\n"); /* split string on newline chars */
   while (tok != NULL)
      {
      /* a type line starts out a new response */
      if ((tmpp = strafter("TYPE:", tok)) != NULL)
         {
         if (strstr(tmpp, "response") == NULL)
            {
            sprintf(messbuf, "This is not a response packet {%s}", data);
            warning(messbuf);
            }
         }
      else if ((tmpp = strafter("ID:", tok)) != NULL)
         {
         strcpy(tmpresponse.id, tmpp);
         }

      else if ((tmpp = strafter("THEME:", tok)) != NULL)
         {
         strcpy(tmpresponse.theme, tmpp);
         }

      else if ((tmpp = strafter("VERSION:", tok)) != NULL)
         {
         tmpresponse.version = atol(tmpp);
         }

       else if ((tmpp = strafter("ADDRESS:", tok)) != NULL)
         {
         strcpy(tmpresponse.address, tmpp);
         }

      else if ((tmpp = strafter("USER:", tok)) != NULL)
         {
         strcpy(tmpresponse.user, tmpp);
         }

      else if ((tmpp = strafter("TIME:", tok)) != NULL)
         {
         tmpresponse.time = atol(tmpp);
         }

      /* handle the multiple responses that are possible */
      else if ((tmpp = strafter("RESPONSE:", tok)) != NULL)
         {
         numr++; totalr++;
         if (totalr > MAX_RESPONSES)
            {
            sprintf(messbuf,
            "Maximum number of responses reached when receiving response packet {%s}", 
            data);
            warning(messbuf);
            return(0);
            }
         strcpy(tmpresponse.response[numr], tmpp);
         }
         
      else
         {
         sprintf(messbuf, 
           "WARNING: did not recognize this line {%s} from packet {%s}",
           tok, data);
         warning(messbuf);
         }
         
      tok = strtok(NULL, "\n"); /* split string again */
      } /* end of while loop */

   /* check that this is not one of my packets */
   remoteip = inet_addr(tmpresponse.address);

   if (remoteip == hostaddr) /* if from local host */
      {
      if (logging)
         {
         logger("response packet is from this host, ignoring\n");
         }
      return(0);
      }
      
   sprintf(messbuf, "IN: %s@%s:response", tmpresponse.user, tmpresponse.address);
   traffic(messbuf);
      
   /* check that this is for the current theme */
   if (strcmp(tmpresponse.theme, theme) != 0)
      {
      sprintf(messbuf, "Response received for other theme: %s", tmpresponse.theme);
      /* warning(messbuf) */;
      traffic(messbuf);
      /* return(0); */
      }
      
   /* check that the response had not expired */
   
   /* get the correct item in the topic structure */
   targettopic = 0;
   for (t = 1; t <= numtopics; t++)
      {
      if (strcmp(tmpresponse.id, topiclist[t].qid) == 0)
         {
         /* t will be the appropriate topic */
         targettopic = t;
         break;
         }
      }
      

   if (! targettopic)
      {
      sprintf(messbuf, "Can't find topic for a response: %s", tmpresponse.theme);
      traffic(messbuf);
      return(0);
      }
         
   if (topiclist[targettopic].decay_time != 0)
      {
      time(&now);
      ntpnow = unixtontp(now);
      tdiff = ntpnow - tmpresponse.time;
      
      /* life time expressed in minutes so multiply by 60 */
      if (tdiff > (topiclist[targettopic].decay_time * DECAY_FACTOR))
         {
         if (logging)
            {
            logger("incoming response has already expired\n");
            }
         return(0);
         }
      }
  
      
   /* check that this is not a duplicate response */
   for (i = 1; i<= numresponses; i++)
      {
      if (
         (strcmp(tmpresponse.id, responselist[i].id) == 0) &&
         (strcmp(tmpresponse.user, responselist[i].user) == 0) &&
         (strcmp(tmpresponse.address, responselist[i].address) == 0)
         ) 
         {
         if (tmpresponse.version == responselist[i].version)
            {
            if (logging)
               {
               logger("Duplicate response ignored\n");
               }
            traffic("...duplicate response ignored");
            return(0);
            }
         else
            {
            /* new response found */
            
            /* update old data structure */
            responselist[i].version = tmpresponse.version;
            responselist[i].time = tmpresponse.version;
            
            for (j = 1; j <= MAX_RESPONSES; j++)
               {
               responselist[i].response[j] = strdup("");
               if (tmpresponse.response[j])
                  {
                  responselist[i].response[j] = strdup(tmpresponse.response[j]);
                  }
               }
               
            if (logging)
               {
               logger("...response replaced\n");
               }
            traffic("...response replaced");
            return(0);
            }
         }
      }

   /* add it to the list of responses and update data */
   numresponses++;
   
   responselist[numresponses].id = strdup(tmpresponse.id);
   responselist[numresponses].theme = strdup(tmpresponse.theme);
   responselist[numresponses].version = tmpresponse.version;
   responselist[numresponses].user = strdup(tmpresponse.user);
   responselist[numresponses].address = strdup(tmpresponse.address);
   for (j = 1; j <= totalr; j++)
      {
      responselist[numresponses].response[j] = strdup(tmpresponse.response[j]);
      }
   responselist[numresponses].time = tmpresponse.time;
   }
   
/* ----------------------------------------------------------------- */
/* parse an incoming member packet */

parse_member(char *data)
   {
   char  *strafter();
   char *tok = NULL;
   char tmpl[1024];
   char *tmpp = tmpl;
   int i;
   time_t now;
   char *iptohost();
   
   /* reset the tmp structure */
   strcpy(tmpmember.theme, "");
   strcpy(tmpmember.user, "");
   strcpy(tmpmember.address, "");
   strcpy(tmpmember.host, "");
   strcpy(tmpmember.os, "");
   strcpy(tmpmember.version, "");
   tmpresponse.time = 0;
   
   strcpy(buff, data); /* set buff to the beginning */
   tok = strtok(buff, "\n"); /* split string on newline chars */
   while (tok != NULL)
      {
      /* a type line starts out a new response */
      if ((tmpp = strafter("TYPE:", tok)) != NULL)
         {
         if (strstr(tmpp, "member") == NULL)
            {
            sprintf(messbuf, "This is not a member packet {%s}", data);
            warning(messbuf);
            }
         }
      else if ((tmpp = strafter("THEME:", tok)) != NULL)
         {
         strcpy(tmpmember.theme, tmpp);
         }

      else if ((tmpp = strafter("USER:", tok)) != NULL)
         {
         strcpy(tmpmember.user, tmpp);
         }
         
       else if ((tmpp = strafter("ADDRESS:", tok)) != NULL)
         {
         strcpy(tmpmember.address, tmpp);
         }

       else if ((tmpp = strafter("OS:", tok)) != NULL)
         {
         strcpy(tmpmember.os, tmpp);
         }

       else if ((tmpp = strafter("VERSION:", tok)) != NULL)
         {
         strcpy(tmpmember.version, tmpp);
         }

      else
         {
         sprintf(messbuf, 
           "WARNING: did not recognize this line {%s} from packet {%s}",
           tok, data);
         warning(messbuf);
         }
         
      tok = strtok(NULL, "\n"); /* split string again */
      } /* end of while loop */

   sprintf(messbuf, "IN: %s@%s:member\n  %s - %s", 
      tmpmember.user, tmpmember.address, tmpmember.version, tmpmember.os);
   traffic(messbuf);
   
   /* check that this is for the current theme */
   if (strcmp(tmpmember.theme, theme) != 0)
      {
      sprintf(messbuf, "Membership received for other theme: %s", 
            tmpmember.theme);
      /* warning(messbuf); */
      traffic(messbuf);
      /* return(0); */
      }

   time(&now);
   
   /* check for duplicate member info */
   for (i = 1; i<= nummembers; i++)
      {
      if (
         (strcmp(tmpmember.user, memberlist[i].user) == 0) &&
         (strcmp(tmpmember.address, memberlist[i].address) == 0)
         ) 
         {
         /* update the time stamp */
         memberlist[i].time = now;
         if (logging)
            {
            logger("Previous membership updated\n");
            }
         traffic("...membership updated");
         return(0);
         }
      }
      
   /* add new member to the list */
               
   nummembers++;
   if (nummembers >= MAX_MEMBERS)
      {
      die("Too many members for storage array\n");
      }
   
   memberlist[nummembers].theme = strdup(tmpmember.theme);
   memberlist[nummembers].user = strdup(tmpmember.user);
   memberlist[nummembers].address = strdup(tmpmember.address);
   memberlist[nummembers].host = strdup(iptohost(tmpmember.address));
   memberlist[nummembers].os = strdup(tmpmember.os);
   memberlist[nummembers].version = strdup(tmpmember.version);
   memberlist[nummembers].time = now;
   
   }

/* ----------------------------------------------------------------- */
/* parse an incoming question packet */

parse_question(char *data)
   {
   char *strafter();
   char *tok = NULL;
   char tmpl[1024];
   char *tmpp = tmpl;
   int i;
   int thistopic;
   unsigned long remoteip;
   int numr, totalr;
   char *remoteadd = NULL;
   char *remoteuser = NULL;
   
   numr = 0; totalr = 0;
   
   /* clean out the tmp space */
   strcpy(tmptopic.qid, "");
   strcpy(tmptopic.theme, "");
   tmptopic.version = 0;
   strcpy(tmptopic.email, "");
   strcpy(tmptopic.summary, "");
   strcpy(tmptopic.question, "");
   strcpy(tmptopic.option, "");
   tmptopic.decay_time = 0;
   tmptopic.remind_time = 0;
   tmptopic.time = 0;
   for (i = 1; i <= MAX_RESPONSES; i++)
      {
      strcpy(tmptopic.response[i], "");
      }
   
   strcpy(buff, data); /* set buff to the beginning */
   tok = strtok(buff, "\n"); /* split string on newline chars */
   while (tok != NULL)
      {
      /* a summary line starts out a new topic */
      if ((tmpp = strafter("TYPE:", tok)) != NULL)
         {
         if (strstr(tmpp, "question") == NULL)
            {
            printf("This is not a question packet {%s}\n", tok);
            warning("problem with parsing packet");
            return(0);
            }
         }
      else if ((tmpp = strafter("QID:", tok)) != NULL)
         {
         strcpy(tmptopic.qid, tmpp);
         }

      else if ((tmpp = strafter("THEME:", tok)) != NULL)
         {
         strcpy(tmptopic.theme, tmpp);
         }

      else if ((tmpp = strafter("VERSION:", tok)) != NULL)
         {
         tmptopic.version = atol(tmpp);
         }

      else if ((tmpp = strafter("EMAIL:", tok)) != NULL)
         {
         strcpy(tmptopic.email, tmpp);
         }

      else if ((tmpp = strafter("SUMMARY:", tok)) != NULL)
         {
         strcpy(tmptopic.summary, tmpp);
         }

      else if ((tmpp = strafter("QUESTION:", tok)) != NULL)
         {
         strcpy(tmptopic.question, tmpp);
         }

      else if ((tmpp = strafter("OPTION:", tok)) != NULL)
         {
         strcpy(tmptopic.option, tmpp);
         }

      else if ((tmpp = strafter("DECAY:", tok)) != NULL)
         {
         tmptopic.decay_time = atoi(tmpp);
         }

      else if ((tmpp = strafter("REMIND:", tok)) != NULL)
         {
         tmptopic.remind_time = atoi(tmpp);
         }

      else if ((tmpp = strafter("TIME:", tok)) != NULL)
         {
         tmptopic.time = atol(tmpp);
         }

      /* handle the multiple responses that are possible */
      else if ((tmpp = strafter("RESPONSE:", tok)) != NULL)
         {
         numr++; totalr++;
         if (totalr > MAX_RESPONSES)
            {
            sprintf(messbuf,
            "Maximum number of responses reached when receiving question packet {%s}", 
            data);
            die(messbuf);
            }
         strcpy(tmptopic.response[numr], tmpp);
         }
         
      else
         {
         sprintf(messbuf, 
           "WARNING: did not recognize this line from packet {%s}", tok);
         warning(messbuf);
         }
         
      tok = strtok(NULL, "\n"); /* split string again */
      } /* end of while loop */

   /* check that this is not one of my packets */
   
   remoteadd = getword(tmptopic.qid, 2);
   remoteip = inet_addr(remoteadd);

   if (remoteip == hostaddr) /* if from local host */
      {
      if (logging)
         {
         logger("question packet is from this host, ignoring\n");
         }
      free(remoteadd);
      return(0);
      }
      
   remoteuser = getword(tmptopic.qid, 1);
   
   sprintf(messbuf, "IN: %s@%s:question", remoteuser, remoteadd);
   traffic(messbuf);
   
   free(remoteadd);
   free(remoteuser);

   /* check that this is for the current theme */
   if (strcmp(tmptopic.theme, theme) != 0)
      {
      sprintf(messbuf, "Question received for other theme: %s", tmptopic.theme);
      traffic(messbuf);
      /* return(0); */
      }

   /* check that this is not a duplicate topic */
   for (i = 1; i<= numtopics; i++)
      {
      if (
         (strcmp(tmptopic.qid, topiclist[i].qid) == 0) &&
         (tmptopic.version == topiclist[i].version)
         ) /* duplicate found */
         {
         if (logging)
            {
            logger("...duplicate question ignored\n");
            }
         traffic("...duplicate question ignored");
         return(0);
         }
      }
      
      

   /* add it to the list of topics and update display */
   numtopics++;
   if (numtopics > MAX_TOPICS)
      {
      die("Maximum number of topics reached.\n");
      }
   thistopic = numtopics;
 
   topiclist[thistopic].qid = strdup(tmptopic.qid);
   topiclist[thistopic].theme = strdup(tmptopic.theme);
   topiclist[thistopic].version = tmptopic.version;
   topiclist[thistopic].email = strdup(tmptopic.email);
   topiclist[thistopic].summary = strdup(tmptopic.summary);
   topiclist[thistopic].question = strdup(tmptopic.question);
   topiclist[thistopic].option = strdup(tmptopic.option);
   topiclist[thistopic].decay_time = tmptopic.decay_time;
   topiclist[thistopic].remind_time = tmptopic.remind_time;
   topiclist[thistopic].time = tmptopic.time;
   for (i = 1; i <= totalr; i++)
      {
      topiclist[thistopic].response[i] = strdup(tmptopic.response[i]);
      }
   
   listtopics();
   }

   
/* ----------------------------------------------------------- */
/* handle an incoming request to delete a question */

parse_delete_question(char *data)
   {
   char *strafter();
   char *tok = NULL;
   char tmpl[1024];
   char *tmpp = tmpl;
   int i, dreq;
   unsigned long remoteip;
   char *origuser = NULL;
   char *origadd = NULL;
   
   /* set the tmp variables to nothing ? */
   strcpy(did, "");
   strcpy(dtheme, "");
   strcpy(daddress, "");
   strcpy(duser, "");
   
       
   strcpy(buff, data); /* set buff to the beginning */
   tok = strtok(buff, "\n"); /* split string on newline chars */

   while (tok != NULL)
      {
      /* a summary line starts out a new topic */
      if ((tmpp = strafter("TYPE:", tok)) != NULL)
         {
         if (strstr(tmpp, "delete_question") == NULL)
            {
            printf("This is not a delete question packet {%s}\n", tok);
            warning("problem with parsing packet");
            return(0);
            }
         }
      else if ((tmpp = strafter("ID:", tok)) != NULL)
         {
         strcpy(did, tmpp);
         }

      else if ((tmpp = strafter("THEME:", tok)) != NULL)
         {
         strcpy(dtheme, tmpp);
         }

      else if ((tmpp = strafter("ADDRESS:", tok)) != NULL)
         {
         strcpy(daddress, tmpp);
         }

      else if ((tmpp = strafter("USER:", tok)) != NULL)
         {
         strcpy(duser, tmpp);
         }

      else
         {
         sprintf(messbuf, 
           "WARNING: did not recognize this line from packet {%s}", tok);
         warning(messbuf);
         }
         
      tok = strtok(NULL, "\n"); /* split string again */
      
      } /* end of while loop */

   /* check that this is not one of my packets */
   remoteip = inet_addr(daddress);

   if (remoteip == hostaddr) /* if from local host */
      {
      if (logging)
         {
         logger("packet is from this host, ignoring\n");
         }
      return(0);
      }
      
   sprintf(messbuf, "IN: %s@%s:delete question request", duser, daddress);
   traffic(messbuf);

   /* check that this is for the current theme */
   if (strcmp(tmptopic.theme, theme) != 0)
      {
      sprintf(messbuf, "Question delete message received for other theme: %s", dtheme);
      traffic(messbuf);
      /* return(0); */
      }
      
   /* find the unwanted question */
   dreq = 0;
   for (i = 1; i<= numtopics; i++)
      {
      if (strcmp(did, topiclist[i].qid) == 0)
         {
         dreq = i;
         }
      }
   if (! dreq)
      {
      sprintf(messbuf, "...cannot find a question to delete");
      traffic(messbuf);
      return(0);
      }
      
   origadd = getword(topiclist[dreq].qid, 2);
   origuser = getword(topiclist[dreq].qid, 1);
   if (
      (strcmp(duser, origuser) != 0) ||
      (strcmp(daddress, origadd) != 0)
      ) /* match not found */
      {
      sprintf(messbuf, 
      "...delete request from %s@%s does not match existing question from %s@%s, ignoring it",
      duser, daddress, origuser, origadd);
      traffic(messbuf);
      free(origadd);
      free(origuser);
      return(0);
      }
   
   
   /* do the deletion */
   
   /* get rid of the local copy */
   /* shuffle all higher topics down one */
   for (i = dreq; i < numtopics; i++)
      {
      memcpy(&topiclist[i], &topiclist[i+1], sizeof(struct topic));
      }
   /* reset the last space to zero */
   memset(&topiclist[numtopics], 0, sizeof(struct topic));
   numtopics--;
      
   traffic("...question deleted");
   
   /* update the display */

   listtopics();
   free(origadd);
   free(origuser);
   }

